

def get_key_from_request_form(request_form):
    serverkey = request_form['username'] + '_' + request_form['database'] + '_' + request_form['scheme'] + '_' + \
        request_form['streamName'] + '_' + request_form['annotator'] + '_' + \
        request_form['sessions'].replace(";", "_")
    return serverkey[:128]    



import threading

from flask import Blueprint, request, jsonify
from nova_server.utils import thread_utils, status_utils, log_utils

extract = Blueprint("extract", __name__)


@extract.route("/extract", methods=["POST"])
def extract_thread():
    if request.method == "POST":
        thread = extract_data(request.form)
        thread_id = thread.name
        status_utils.add_new_job(thread_id, extract.name)
        data = {"job_id": thread_id}
        thread.start()
        return jsonify(data)


@thread_utils.ml_thread_wrapper
def extract_data(request_form):
    def update_progress(msg):
        status_utils.update_progress(threading.current_thread().name, msg)

    # Init logging
    status_utils.update_status(threading.current_thread().name, status_utils.JobStatus.RUNNING)

    update_progress('Initalizing')
    logger = log_utils.get_logger_for_thread(__name__)
    log_conform_request = dict(request_form)
    log_conform_request['password'] = '---'
    logger.info(f"Start Extracting with request {log_conform_request}")

    logger.error('Not implemented.')

    update_progress('Done')
    status_utils.update_status(threading.current_thread().name, status_utils.JobStatus.ERROR)
    return




import xml.etree.ElementTree as ET
from pathlib import Path


class Trainer:
    def __init__(self,
                 model_script_path: str = '',
                 model_option_path: str = '',
                 model_option_string: str = '',
                 model_weights_path: str = '',
                 model_stream: int = 0,
                 model_create: str = "PythonModel",
                 users: list = None,
                 classes: list = None,
                 streams: list = None,
                 register: list = None,
                 info_trained: bool = False,
                 meta_right_ctx: int = 0,
                 meta_left_ctx: int = 0,
                 meta_balance: str = 'none',
                 meta_backend: str = 'Python',
                 ssi_v="5",
                 xml_version="1.0"
                 ):

        self.model_script_path = model_script_path
        self.model_option_path = model_option_path
        self.model_optstr = model_option_string
        self.model_weights_path = model_weights_path
        self.model_stream = model_stream
        self.model_create = model_create
        self.users = users if users is not None else []
        self.classes = classes if classes is not None else []
        self.streams = streams if streams is not None else []
        self.register = register if register is not None else []
        self.info_trained = info_trained
        self.meta_right_ctx = meta_right_ctx
        self.meta_left_ctx = meta_left_ctx
        self.meta_balance = meta_balance
        self.meta_backend = meta_backend
        self.ssi_v = ssi_v
        self.xml_version = xml_version

    def load_from_file(self, fp):
        root = ET.parse(Path(fp))
        info = root.find('info')
        meta = root.find('meta')
        register = root.find('register')
        streams = root.find('streams')
        classes = root.find('classes')
        users = root.find('users')
        model = root.find('model')

        if info is not None:
            self.info_trained = info.get('trained')
        if meta is not None:
            self.meta_left_ctx = int(meta.get('leftContext', default='0'))
            self.meta_right_ctx = int(meta.get('rightContext', default='0'))
            self.meta_balance = meta.get('balance', default='none')
            self.meta_backend = meta.get('backend', default='Python')
        if register is not None:
            for r in register:
                self.register.append(r.attrib)
        if streams is not None:
            for s in streams:
                self.streams.append(s.attrib)
        if classes is not None:
            for c in classes:
                self.classes.append(c.attrib)
        if users is not None:
            for u in users:
                self.users.append(u.attrib)
        if model is not None:
            self.model_stream = model.get('stream', default='0')
            self.model_create = model.get('create', default='PythonModel')
            self.model_option_path = model.get('option', default='')
            self.model_script_path = model.get('script', default='')
            self.model_weights_path = model.get('path', default='')
            self.model_optstr = model.get('optstr', default='')

    def write_trainer_to_file(self, fp):
        root = ET.Element('trainer')
        ET.SubElement(root, 'info', trained=str(self.info_trained))
        ET.SubElement(root, 'meta', leftContext=str(self.meta_left_ctx), rightContex=str(self.meta_right_ctx), balance=self.meta_balance, backend=self.meta_backend)
        register = ET.SubElement(root, 'register')
        for r in self.register:
            ET.SubElement(register, 'item', **r)
        streams = ET.SubElement(root, 'streams')
        for s in self.streams:
            ET.SubElement(streams, 'item', **s)
        classes = ET.SubElement(root, 'classes')
        for c in self.classes:
            ET.SubElement(classes, 'item', **c)
        users = ET.SubElement(root, 'users')
        for u in self.users:
            ET.SubElement(users, 'item', **u)
        ET.SubElement(root, 'model', create=self.model_create, stream=str(self.model_stream), path=self.model_weights_path, script=self.model_script_path, optstr=self.model_optstr, option=self.model_option_path)

        tree = ET.ElementTree(root)
        ET.indent(tree, space="    ", level=0)

        if not fp.suffix:
            fp = fp.with_suffix('.trainer')
        tree.write(fp)


if __name__ == '__main__':
    trainer_in_fp = Path(
        r'Z:\nova\cml\models\trainer\discrete\base_emotions\feature{compare[480ms,40ms,480ms]}\linsvm\linsvm.compare.trainer')
    trainer_out_fp = Path('.test_trainer.trainer')

    trainer = Trainer()
    trainer.load_from_file(trainer_in_fp)
    trainer.write_trainer_to_file(trainer_out_fp)




import sys
import bson
import copy
import warnings
import numpy as np
import nova_server.utils.path_config as cfg

sys.path.append(cfg.hcai_datasets_dir)
from hcai_datasets.hcai_nova_dynamic.nova_db_handler import NovaDBHandler
from hcai_datasets.hcai_nova_dynamic.utils import nova_data_utils

MAX_MONGO_DB_DOC_SIZE = 16777216
database = None
scheme = None
session = None
annotator = None
roles = None


# Format of results must be: {values: [value1, value2, value3, ...], confidences: [conf1, conf2, conf3, ...], start_frame: x, ...}
def check_format(variable, logger):
    text = ""
    if type(variable) is not dict:
        text = 'Result type must be a dictionary!'
    if "values" not in variable or "confidences" not in variable:
        text = 'Result type must contain "values" and "confidences" as keys.'
    if not type(variable["values"]) is list or not type(variable["confidences"]):
        text = 'Results["values"] and Results["confidences"] must respectively be a list.'

    if text != "":
        logger.info(text)
        raise Exception(text)


def write_annotation_to_db(request_form, results: dict, logger):
    global database, scheme, session, annotator, roles
    check_format(results, logger)

    # TODO check if we really need to establish a new connection to the database
    db_config_dict = {
        'ip': request_form["server"].split(':')[0],
        'port': int(request_form["server"].split(':')[1]),
        'user': request_form["username"],
        'password': request_form["password"]
    }

    db_handler = NovaDBHandler(db_config_dict=db_config_dict)
    database = request_form['database']
    scheme = request_form['scheme']
    session = request_form['sessions']
    annotator = request_form['annotator']
    roles = request_form['roles']

    if request_form["schemeType"] == "DISCRETE":
        write_discrete_to_db(request_form, results, db_handler, logger)
    elif request_form["schemeType"] == "FREE":
        write_freeform_to_db(request_form, results, db_handler, logger)
    elif request_form["schemeType"] == "CONTINUOUS":
        write_continuous_to_db(request_form, results, db_handler, logger)
    elif request_form["schemeType"] == "POINT":
        pass  # todo
    elif request_form["schemeType"] == "DISCRETE_POLYGON" or request_form["schemeType"] == "POLYGON":
        write_polygons_to_db(request_form, results, db_handler, logger)


def write_freeform_to_db(request_form, results: dict, db_handler, logger):
    annotations = {}
    """Temp fix"""
    results.pop('values', None)
    results.pop('confidences', None)

    for frame, results in results.items():
        frame_info = frame.split('_')
        frame_from = float(frame_info[-2])
        frame_to = float(frame_info[-1])
        for stream_id, anno in results.items():

            conf = anno['conf']
            name = anno['name']

            if stream_id not in annotations.keys():
                annotations[stream_id] = []

            annotations[stream_id].append({
                "from": frame_from,
                "to": frame_to,
                "conf": conf,
                "name": name
            })

    for anno_id, anno in annotations.items():
        # TODO does not work with flattened roles
        role, stream = anno_id.split('.')

        db_handler.set_annos(
            database=database,
            scheme=scheme,
            session=session,
            annotator=annotator,
            role=role,
            annos=anno,
        )


def write_discrete_to_db(request_form, results: dict, db_handler, logger):
    # TODO: We only take one role into account in this case. Fix
    #role = roles.split(';')[0]
    role = results['roles']
    frame_size = nova_data_utils.parse_time_string_to_ms(request_form['frameSize'])
    mongo_scheme = db_handler.get_mongo_scheme(scheme, database)
    annos = []
    start_frame = 0

    if request_form['startTime'] != '0':
        annos_db = db_handler.get_annos(dataset=database, scheme=scheme, session=session, annotator=annotator,
                                        roles=role)

        start_frame = float(request_form['startTime']) / 1000
        annos = list(filter(lambda x: float(x['from']) < start_frame, annos_db))

    last_label = None
    current_label_start = 1

    values = results["values"]

    for id, value in enumerate(values):
        # current label is different from the one before
        if not value == last_label and last_label is not None:
            frame_from = str(start_frame + ((current_label_start * frame_size) / 1000.0))
            frame_to = str(start_frame + ((id * frame_size) / 1000.0))
            if last_label < len(mongo_scheme[0]['labels']):
                annos.append({
                    'from': frame_from,
                    'to': frame_to,
                    'conf': results["confidences"][id],
                    'id': int(last_label)
                })
            current_label_start = id
        last_label = value

    db_handler.set_annos(
        database=database,
        scheme=scheme,
        session=session,
        annotator=annotator,
        role=role,
        annos=annos,
    )


def write_continuous_to_db(request_form, results: dict, db_handler, logger):
    role = results['roles']
    #role = roles.split(';')[0]
    annos = []

    if request_form['startTime'] != '0':
        annos_db = db_handler.get_annos(dataset=database, scheme=scheme, session=session, annotator=annotator,
                                        roles=roles)

        start_frame = int(int(request_form['startTime']) / int(request_form['frameSize'][:-2]) + 1)
        annos = annos_db[:start_frame]

    #normalized = (results["values"] - min(results["values"])) / (max(results["values"]) - min(results["values"]))

    for value, confidence in zip(results["values"], results["confidences"]):
        annos.append({
            'score': value,
            'conf': confidence
        })

    db_handler.set_annos(
        database=database,
        scheme=scheme,
        session=session,
        annotator=annotator,
        role=role,
        annos=annos,
    )


def write_polygons_to_db(request_form, results: dict, db_handler, logger):
    mongo_scheme = db_handler.get_mongo_scheme(scheme, database)
    mongo_annotator = db_handler.get_mongo_annotator(annotator, database)
    mongo_role = db_handler.get_mongo_role(roles, database)
    mongo_session = db_handler.get_mongo_session(session, database)

    mongo_annotations = db_handler.get_annotation_docs(
        mongo_scheme,
        mongo_session,
        mongo_annotator,
        mongo_role,
        database,
        db_handler.ANNOTATION_COLLECTION,
    )

    start_frame = int(float(results["start_frame"])) - 1

    mongo_data_id = None
    data_backup_id = None
    if mongo_annotations:
        if mongo_annotations[0]["isLocked"]:
            warnings.warn(f"Can't overwrite locked annotation {str(mongo_annotations[0]['_id'])}")
            return ""
        else:
            mongo_data_id = mongo_annotations[0]["data_id"]
            data_backup_id = mongo_annotations[0]["data_backup_id"]
    logger.info("...fetch documents...")
    # 1. Get the doc (will get merged if other docs are there are nextEntry ID's)
    main_docs = db_handler.get_data_docs_by_prop(mongo_data_id, "_id", database)
    backup_doc = db_handler.get_docs_by_prop(mongo_data_id, "_id", database, db_handler.ANNOTATION_DATA_COLLECTION)[0]
    logger.info("...fill documents...")
    # 2. Fill the doc with the Predictions
    main_docs = update_polygon_doc(main_docs, results["values"], results["confidences"], start_frame)

    # 3. Check if the doc is too large (if, separate it)
    if len(bson.BSON.encode(main_docs)) >= MAX_MONGO_DB_DOC_SIZE:
        main_docs = separate_doc(main_docs)
    if not isinstance(main_docs, list):
        main_docs = [main_docs]
    main_docs[0]['_id'] = mongo_data_id

    logger.info("...separate docs if necessary...")
    # 4 Delete old back-up (with tail-docs)
    db_handler.delete_doc_with_tail(data_backup_id, database)

    logger.info("...upload docs...")
    # 5. Update the backup ID
    backup_doc['_id'] = data_backup_id
    db_handler.insert_doc_by_prop(backup_doc, database, db_handler.ANNOTATION_DATA_COLLECTION)
    db_handler.delete_doc_by_prop(mongo_data_id, "_id", database, db_handler.ANNOTATION_DATA_COLLECTION)

    # 6. Upload the doc(s)
    for doc in main_docs:
        db_handler.insert_doc_by_prop(doc, database, db_handler.ANNOTATION_DATA_COLLECTION)


def separate_doc(doc) -> list:
    left = copy.deepcopy(doc)
    left["_id"] = bson.objectid.ObjectId()
    right = copy.deepcopy(doc)
    right["_id"] = bson.objectid.ObjectId()
    half_count = int(len(doc['labels']) / 2)
    left['labels'] = left['labels'][0:half_count]
    right['labels'] = right['labels'][half_count:]

    result_list = []
    if len(bson.BSON.encode(left)) >= MAX_MONGO_DB_DOC_SIZE:
        result_list += separate_doc(left)
    else:
        result_list.append(left)

    if len(bson.BSON.encode(right)) >= MAX_MONGO_DB_DOC_SIZE:
        result_list += separate_doc(right)
    else:
        result_list.append(right)

    return result_list


def update_polygon_doc(data_doc, polygons, confidences, start_frame):
    current_frame = -1
    for frame_id, frame in enumerate(data_doc['labels']):
        if frame_id >= start_frame + len(polygons):
            return data_doc
        if frame_id >= start_frame:
            polygons_per_frame = polygons[frame_id - start_frame]
            for label_id, polygons_per_label_type in enumerate(polygons_per_frame):
                for polygon_id, polygon in enumerate(polygons_per_label_type):
                    label = label_id + 1
                    points_for_db = []
                    points = np.reshape(polygon, newshape=[int(polygon.shape[0] / 2), 2])
                    for point in points:
                        points_for_db.append({'x': int(point[0]), 'y': int(point[1])})

                    # delete the content of the current frame, the new prediction values have to be set
                    if current_frame != frame_id:
                        current_frame = frame_id
                        data_doc['labels'][frame_id]['polygons'] = []

                    data_doc['labels'][frame_id]['polygons'].append(
                        {'label': label,
                         'confidence': round(confidences[frame_id - start_frame][label_id][polygon_id], 2),
                         'points': points_for_db})

    return data_doc



import tensorflow as tf
import io
import numpy as np
import io as inputoutput
import base64
import json
import pickle
import warnings
import site as s
from tensorflow.keras.models import load_model
from PIL import Image
from PIL import Image as pilimage
from flask import Blueprint, Response, request
from lime import lime_image
from lime import lime_tabular
from skimage.segmentation import mark_boundaries

from nova_server.utils import explain_utils, img_utils

# Initialization
s.getusersitepackages()
warnings.simplefilter("ignore")
graph = tf.compat.v1.get_default_graph()
explain = Blueprint("explain", __name__)


@explain.route("/tfexplain", methods=["POST"])
def explain_tfexplain():
    global graph
    data = {"success": "failed"}
    #      # ensure an image was properly uploaded to our endpoint
    if request.method == "POST":
        if request.form.get("image"):

            explainer = request.args.get("explainer")
            # with graph.as_default():
            model_path = request.form.get("model_path")
            model = load_model(model_path)

            # read the image in PIL format
            image64 = request.form.get("image")
            image = base64.b64decode(image64)
            image = Image.open(io.BytesIO(image))
            image = img_utils.prepare_image(image, target=(224, 224))
            image = image * (1.0 / 255)
            # img = tf.keras.preprocessing.image.img_to_array(image)
            prediction = model.predict(image)
            topClass = explain_utils.getTopXpredictions(prediction, 1)
            print(topClass[0])
            image = np.squeeze(image)

            if explainer == "GRADCAM":
                im = ([image], None)
                from tf_explain.core.grad_cam import GradCAM

                exp = GradCAM()
                imgFinal = exp.explain(im, model, class_index=topClass[0][0])
                # exp.save(imgFinal, ".", "grad_cam.png")

            elif explainer == "OCCLUSIONSENSITIVITY":
                im = ([image], None)
                from tf_explain.core.occlusion_sensitivity import OcclusionSensitivity

                exp = OcclusionSensitivity()
                imgFinal = exp.explain(
                    im, model, class_index=topClass[0][0], patch_size=10
                )
                # exp.save(imgFinal, ".", "grad_cam.png")

            elif explainer == "GRADIENTSINPUTS":
                im = (np.array([image]), None)
                from tf_explain.core.gradients_inputs import GradientsInputs

                exp = GradientsInputs()
                imgFinal = exp.explain(im, model, class_index=topClass[0][0])
                # exp.save(imgFinal, ".", "gradients_inputs.png")

            elif explainer == "VANILLAGRADIENTS":
                im = (np.array([image]), None)
                from tf_explain.core.vanilla_gradients import VanillaGradients

                exp = VanillaGradients()
                imgFinal = exp.explain(im, model, class_index=topClass[0][0])
                # exp.save(imgFinal, ".", "gradients_inputs.png")

            elif explainer == "SMOOTHGRAD":
                im = (np.array([image]), None)
                from tf_explain.core.smoothgrad import SmoothGrad

                exp = SmoothGrad()
                imgFinal = exp.explain(im, model, class_index=topClass[0][0])
                # exp.save(imgFinal, ".", "gradients_inputs.png")

            elif explainer == "INTEGRATEDGRADIENTS":
                im = (np.array([image]), None)
                from tf_explain.core.integrated_gradients import IntegratedGradients

                exp = IntegratedGradients()
                imgFinal = exp.explain(im, model, class_index=topClass[0][0])
                # exp.save(imgFinal, ".", "gradients_inputs.png")

            elif explainer == "ACTIVATIONVISUALIZATION":
                # need some solution to find out and submit layers name
                im = (np.array([image]), None)
                from tf_explain.core.activations import ExtractActivations

                exp = ExtractActivations()
                imgFinal = exp.explain(im, model, layers_name=["activation_1"])
                # exp.save(imgFinal, ".", "gradients_inputs.png")

            img = pilimage.fromarray(imgFinal)
            imgByteArr = inputoutput.BytesIO()
            img.save(imgByteArr, format="JPEG")
            imgByteArr = imgByteArr.getvalue()

            img64 = base64.b64encode(imgByteArr)
            img64_string = img64.decode("utf-8")

            data["explanation"] = img64_string
            data["prediction"] = str(topClass[0][0])
            data["prediction_score"] = str(topClass[0][1])
            data["success"] = "success"

    return Response(json.dumps(data), mimetype="text/plain")


@explain.route("/innvestigate", methods=["POST"])
def explain_innvestigate():
    global graph
    data = {"success": "failed"}
    # ensure an image was properly uploaded to our endpoint
    if request.method == "POST":
        if request.form.get("image"):

            postprocess = request.args.get("postprocess")
            explainer = request.args.get("explainer")
            lrpalpha = float(request.args.get("lrpalpha"))
            lrpbeta = float(request.args.get("lrpbeta"))

            with graph.as_default():
                model_path = request.form.get("model_path")
                model = load_model(model_path)

                # read the image in PIL format
                image64 = request.form.get("image")
                image = base64.b64decode(image64)
                image = Image.open(io.BytesIO(image))

                # preprocess the image and prepare it for classification
                image = img_utils.prepare_image(image, target=(224, 224))
                image = image * (1.0 / 255)

                # print(model.summary())
                model_wo_sm = iutils.keras.graph.model_wo_softmax(model)

                prediction = model.predict(image)
                topClass = explain_utils.getTopXpredictions(prediction, 1)

                model_wo_sm = iutils.keras.graph.model_wo_softmax(model)

                analyzer = []

                if explainer == "GUIDEDBACKPROP":
                    analyzer = innvestigate.analyzer.GuidedBackprop(model_wo_sm)
                elif explainer == "GRADIENT":
                    analyzer = innvestigate.analyzer.Gradient(model_wo_sm)
                elif explainer == "DECONVNET":
                    analyzer = innvestigate.analyzer.Deconvnet(model_wo_sm)
                elif explainer == "LRPEPSILON":
                    analyzer = innvestigate.analyzer.LRPEpsilon(model_wo_sm)
                elif explainer == "LRPZ":
                    analyzer = innvestigate.analyzer.LRPZ(model_wo_sm)
                elif explainer == "LRPALPHABETA":
                    analyzer = innvestigate.analyzer.LRPAlphaBeta(
                        model_wo_sm, alpha=lrpalpha, beta=lrpbeta
                    )
                elif explainer == "DEEPTAYLOR":
                    analyzer = innvestigate.analyzer.DeepTaylor(model_wo_sm)

                # Applying the analyzer
                analysis = analyzer.analyze(image)

                imgFinal = []

                if postprocess == "GRAYMAP":
                    imgFinal = explain_utils.graymap(analysis)[0]
                elif postprocess == "HEATMAP":
                    imgFinal = explain_utils.heatmap(analysis)[0]
                elif postprocess == "BK_PROJ":
                    imgFinal = explain_utils.bk_proj(analysis)[0]
                elif postprocess == "GNUPLOT2":
                    imgFinal = explain_utils.heatmapgnuplot2(analysis)[0]
                elif postprocess == "CMRMAP":
                    imgFinal = explain_utils.heatmapCMRmap(analysis)[0]
                elif postprocess == "NIPY_SPECTRAL":
                    imgFinal = explain_utils.heatmapnipy_spectral(analysis)[0]
                elif postprocess == "RAINBOW":
                    imgFinal = explain_utils.heatmap_rainbow(analysis)[0]
                elif postprocess == "INFERNO":
                    imgFinal = explain_utils.heatmap_inferno(analysis)[0]
                elif postprocess == "GIST_HEAT":
                    imgFinal = explain_utils.heatmap_gist_heat(analysis)[0]
                elif postprocess == "VIRIDIS":
                    imgFinal = explain_utils.heatmap_viridis(analysis)[0]

                imgFinal = np.uint8(imgFinal * 255)

                img = pilimage.fromarray(imgFinal)
                imgByteArr = inputoutput.BytesIO()
                img.save(imgByteArr, format="JPEG")
                imgByteArr = imgByteArr.getvalue()

                img64 = base64.b64encode(imgByteArr)
                img64_string = img64.decode("utf-8")
                data["explanation"] = img64_string
                data["prediction"] = str(topClass[0][0])
                data["prediction_score"] = str(topClass[0][1])
                data["success"] = "success"

    return Response(json.dumps(data), mimetype="text/plain")


@explain.route("/lime", methods=["POST"])
def explain_lime():
    global graph
    data = {"success": "failed"}

    if request.method == "POST":
        if request.form.get("image"):

            top_labels = int(request.args.get("toplabels"))
            hide_color = str(request.args.get("hidecolor"))
            num_samples = int(request.args.get("numsamples"))
            positive_only = str(request.args.get("positiveonly"))
            num_features = int(request.args.get("numfeatures"))
            hide_rest = str(request.args.get("hiderest"))

            # read the image in PIL format
            image64 = request.form.get("image")
            image = base64.b64decode(image64)
            image = Image.open(io.BytesIO(image))

            with graph.as_default():

                model_path = request.form.get("model_path")
                model = load_model(model_path)

                img = img_utils.prepare_image(image, (224, 224))
                img = img * (1.0 / 255)
                prediction = model.predict(img)
                explainer = lime_image.LimeImageExplainer()
                img = np.squeeze(img).astype("double")
                explanation = explainer.explain_instance(
                    img,
                    model.predict,
                    top_labels=top_labels,
                    hide_color=hide_color == "True",
                    num_samples=num_samples,
                )

                top_classes = explain_utils.getTopXpredictions(prediction, top_labels)

                explanations = []

                for cl in top_classes:
                    temp, mask = explanation.get_image_and_mask(
                        cl[0],
                        positive_only=positive_only == "True",
                        num_features=num_features,
                        hide_rest=hide_rest == "True",
                    )
                    img_explained = mark_boundaries(temp, mask)
                    img = Image.fromarray(np.uint8(img_explained * 255))
                    img_byteArr = io.BytesIO()
                    img.save(img_byteArr, format="JPEG")
                    img_byteArr = img_byteArr.getvalue()
                    img64 = base64.b64encode(img_byteArr)
                    img64_string = img64.decode("utf-8")

                    explanations.append((str(cl[0]), str(cl[1]), img64_string))

                data["explanations"] = explanations
                data["success"] = "success"

    return Response(json.dumps(data), mimetype="text/plain")


@explain.route("/tabular", methods=["POST"])
def explain_tabular():
    data = {"success": "failed"}

    # TODO send sample to be explained

    if request.method == "POST":

        if request.form:

            # data_dict = ast.literal_eval(json.loads(flask.request.data))

            print("try open model")
            with open(request.form.get("model_path"), "rb") as f:
                model = pickle.load(f)

            train_data = json.loads(request.form.get("data"))
            dim = json.loads(request.form.get("dim"))
            train_data = np.asarray(train_data)
            train_data = train_data.reshape(((int)(train_data.size / dim), dim))
            sample = json.loads(request.form.get("sample"))

            num_features = int(request.args.get("numfeatures"))

            explainer = lime_tabular.LimeTabularExplainer(
                train_data, mode="classification", discretize_continuous=True
            )
            exp = explainer.explain_instance(
                np.asarray(sample),
                model.predict_proba,
                num_features=num_features,
                top_labels=1,
            )

            explanation_dictionary = {}

            for entry in exp.as_list():
                explanation_dictionary.update({entry[0]: entry[1]})

            data["explanation"] = explanation_dictionary
            data["success"] = "success"

    return Response(json.dumps(data), mimetype="text/plain")



import sys
import nova_server.utils.path_config as cfg

sys.path.append(cfg.hcai_datasets_dir)
from hcai_datasets.hcai_nova_dynamic.hcai_nova_dynamic_iterable import HcaiNovaDynamicIterable
from nova_server.utils.path_config import data_dir

def dataset_from_request_form(request_form):
    """
    Creates a tensorflow dataset from nova dynamically
    :param request_form: the requestform that specifices the parameters of the dataset
    """
    db_config_dict = {
        'ip': request_form["server"].split(':')[0],
        'port': int(request_form["server"].split(':')[1]),
        'user': request_form["username"],
        'password': request_form["password"]
    }

    flattenSamples = False
    if request_form["flattenSamples"] == "true":
        flattenSamples = True

    ds_iter = HcaiNovaDynamicIterable(
        # Database Config
        db_config_path=None,  # os.path.join(os.path.dirname(os.path.abspath(__file__)), 'db.cfg'),
        db_config_dict=db_config_dict,

        # Dataset Config
        dataset=request_form["database"],
        #nova_data_dir=request_form["dataPath"],
        nova_data_dir=data_dir,
        sessions=request_form["sessions"].split(';'),
        roles=request_form["roles"].split(';'),
        schemes=request_form["scheme"].split(';'),
        annotator=request_form["annotator"],
        data_streams=request_form["streamName"].split(' '),

        # Sample Config
        frame_size=request_form["frameSize"],
        left_context=request_form["leftContext"],
        right_context=request_form["rightContext"],
        start=request_form["startTime"],
        end=request_form["endTime"],

        #TODO: This does not work with pytorch bridge when set to true because the data field does not contain the role anymore<.
        # transformation cannot be applied. fix it!
        flatten_samples=flattenSamples,
        supervised_keys=[request_form["streamName"].split(' ')[0],
                         request_form["scheme"].split(';')[0]],

        # Additional Config
        clear_cache=True,
    )

    return ds_iter



import os
import importlib.util

#import torch.cuda

import nova_server.utils.path_config as cfg

from pathlib import Path
from nova_server.utils import db_utils
from flask import Blueprint, request, jsonify
from nova_server.utils.ssi_utils import Trainer
from importlib.machinery import SourceFileLoader
from nova_server.utils.thread_utils import THREADS
from nova_server.utils.status_utils import update_progress
from nova_server.utils.key_utils import get_key_from_request_form
from nova_server.utils import thread_utils, status_utils, log_utils, dataset_utils

predict = Blueprint("predict", __name__)


@predict.route("/predict", methods=["POST"])
def predict_thread():
    if request.method == "POST":
        request_form = request.form.to_dict()
        key = get_key_from_request_form(request_form)
        thread = predict_data(request_form)
        status_utils.add_new_job(key)
        data = {"success": "true"}
        thread.start()
        THREADS[key] = thread
        return jsonify(data)


@thread_utils.ml_thread_wrapper
def predict_data(request_form):
    key = get_key_from_request_form(request_form)
    logger = log_utils.get_logger_for_thread(key)

    # try:
    logger.info("Action 'Predict' started.")
    status_utils.update_status(key, status_utils.JobStatus.RUNNING)
    sessions = request_form["sessions"].split(";")
    roles = request_form["roles"].split(";")
    trainer_file_path = Path(cfg.cml_dir + request_form["trainerFilePath"])
    trainer = Trainer()

    if not trainer_file_path.is_file():
        logger.error("Trainer file not available!")
        status_utils.update_status(key, status_utils.JobStatus.ERROR)
        return None
    else:
        trainer.load_from_file(trainer_file_path)
        logger.info("Trainer successfully loaded.")

    if not trainer.model_script_path:
        logger.error('Trainer has no attribute "script" in model tag.')
        status_utils.update_status(key, status_utils.JobStatus.ERROR)
        return None

    model_script = None

    # Load Data
    for session in sessions:
        request_form["sessions"] = session  # overwrite so we handle each session seperatly..
        for role in roles:
            try:
                update_progress(key, 'Data loading')
                request_form["roles"] = role
                ds_iter = dataset_utils.dataset_from_request_form(request_form)
                logger.info("Prediction data successfully loaded.")
            except ValueError:
                log_utils.remove_log_from_dict(key)
                logger.error("Not able to load the data from the database!")
                status_utils.update_status(key, status_utils.JobStatus.ERROR)
                return

            # Load the model_script only once
            if model_script is None:
                # Load Trainer
                model_script_path = trainer_file_path.parent / trainer.model_script_path
                source = SourceFileLoader("model_script", str(model_script_path)).load_module()
                model_script = source.TrainerClass(ds_iter, logger, request_form)
                # Set Options
                logger.info("Setting options...")
                if not request_form["OptStr"] == '':
                    for k, v in dict(option.split("=") for option in request_form["OptStr"].split(";")).items():
                        if v in ('True', 'False'):
                            model_script.OPTIONS[k] = True if v == 'True' else False
                        elif v == 'None':
                            model_script.OPTIONS[k] = True if v == 'True' else False
                        else:
                            model_script.OPTIONS[k] = v
                        logger.info(k + '=' + v)
                logger.info("...done.")

                # Load Model
                model_weight_path = trainer_file_path.parent / trainer.model_weights_path
                logger.info("Loading model.")
                model_script.load(model_weight_path)
                logger.info("Model loaded.")

            # Load the model only one time as well

            model_script.ds_iter = ds_iter
            model_script.request_form["sessions"] = session
            model_script.request_form["roles"] = role


            logger.info("Execute preprocessing.")
            model_script.preprocess()
            logger.info("Preprocessing done.")

            logger.info("Execute prediction.")
            model_script.predict()
            logger.info("Prediction done.")

            logger.info("Execute postprocessing.")
            results = model_script.postprocess()
            logger.info("Postprocessing done.")

            logger.info("Execute saving process.")
            db_utils.write_annotation_to_db(request_form, results, logger)
            logger.info("Saving process done.")

            # 5. In CML case, delete temporary files..
        if request_form["deleteFiles"] == "True":
            logger.info('Deleting temporary CML files...')
            out_dir = Path(cfg.cml_dir + request_form["trainerOutputDirectory"])
            trainer_name = request_form["trainerName"]
            os.remove(out_dir / trainer.model_weights_path)
            os.remove(out_dir / trainer.model_script_path)
            for f in model_script.DEPENDENCIES:
                os.remove(trainer_file_path.parent / f)
            trainer_fullname = trainer_name + ".trainer"
            os.remove(out_dir / trainer_fullname)
            logger.info('...done')

        logger.info('Prediction completed!')
        status_utils.update_status(key, status_utils.JobStatus.FINISHED)


   # except Exception as e:
    #logger.error('Error:' + str(e))
     #   status_utils.update_status(key, status_utils.JobStatus.ERROR)
    #finally:
    #    del results, ds_iter, ds_iter_pp, model, model_script, model_script_path, model_weight_path, spec



from tensorflow.keras.preprocessing import image as keras_image
import numpy as np

def prepare_image(image, target):
    # if the image mode is not RGB, convert it
    if image.mode != "RGB":
        image = image.convert("RGB")
    # resize the input image and preprocess it
    image = image.resize(target)
    image = keras_image.img_to_array(image)
    image = np.expand_dims(image, axis=0)

    # return the processed image
    return image




class TrainerClass:
    """Includes all the necessary files to run this script"""

    def __init__(self, ds_iter, logger, request_form=None):
        self.model = None
        self.ds_iter = ds_iter
        self.logger = logger
        self.data = None
        self.predictions = None
        self.DEPENDENCIES = []
        self.OPTIONS = {}
        self.request_form = request_form

    def preprocess(self):
        """Possible pre-processing of the data."""
        pass

    def train(self):
        """Trains a model with the given data."""
        pass

    def predict(self):
        """Predicts the given data with the given model. Stores a list with the predicted values to self.predictions."""
        pass

    def postprocess(self) -> list:
        """Possible pro-processing of the data. Returns a list with the pro-processed data."""
        return self.predictions

    def save(self, path) -> str:
        """Stores the weights of the given model at the given path. Returns the path of the weights."""
        modelpath = "pathtofile"
        return modelpath

    def load(self, path):
        """Loads a model with the given path. Returns this model."""
        pass
